# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
__init__ of the module containing the analysis tools
"""

from .equivalent_equation import EquivalentEquation
from .stability import Stability